-- =====================================================================
-- Copyright  2010-2011 by Cryptographic Engineering Research Group (CERG),
-- ECE Department, George Mason University
-- Fairfax, VA, U.S.A.
-- =====================================================================

library ieee;
use ieee.std_logic_1164.all;   
use ieee.numeric_std.all;
use ieee.std_logic_unsigned.all; 
use work.sha3_pkg.all;
use work.groestl_pkg.all;

-- possible generics values: hs = {GROESTL_DATA_SIZE_SMALL, GROESTL_DATA_SIZE_BIG}
-- rom_style = {DISTRIBUTED, COMBINATIONAL}
-- all combinations are allowed 

entity groestl_pq_parallel_folded is
generic (n:integer := GROESTL_DATA_SIZE_SMALL; rom_style : integer := DISTRIBUTED);
port( 
	clk				: in std_logic;
	rst				: in std_logic;
	sel_rd			: in std_logic_vector(2 downto 0);	
	round			: in std_logic_vector(7 downto 0);
	input 			: in std_logic_vector(n-1 downto 0);
   	output 			: out std_logic_vector(n-1 downto 0));
end groestl_pq_parallel_folded;
     

-- combinational implementiation of PQ function - there is no pipelining register 

architecture p_folded_fx2 of groestl_pq_parallel_folded is

signal	after_subbyte		: std_logic_vector(n/2-1 downto 0);
signal	addcons			: std_logic_vector(n-1 downto 0);
signal	after_reg		: std_logic_vector(n/2-1 downto 0);
signal to_sbox			: std_logic_vector(n/2-1 downto 0);
signal	after_shiftrow	: std_logic_vector(n/2-1 downto 0);
constant zero			: std_logic_vector(n/2-1 downto 0):=(others=>'0');
signal after_mixcolumn	 : std_logic_vector(n/2-1 downto 0);
type matrix is array (0 to n/8) of std_logic_vector(7 downto 0);
signal cons_matrix	: matrix;

begin
	
	
gen256: if n=GROESTL_DATA_SIZE_SMALL generate

	addcons(511 downto 504) <= (input(511 downto 504) xor round xor x"00"); 
	addcons(503 downto 456) <= input(503 downto 456); 
	addcons(455 downto 448) <= input(455 downto 448); 
	
	addcons(447 downto 440) <= (input(447 downto 440) xor round xor x"10");
	addcons(439 downto 392) <= input(439 downto 392); 
	addcons(391 downto 384) <= input(391 downto 384); 

	addcons(383 downto 376) <= (input(383 downto 376) xor round xor x"20"); 
	addcons(375 downto 328) <= input(375 downto 328);  
	addcons(327 downto 320) <= input(327 downto 320); 

	addcons(319 downto 312) <= (input(319 downto 312) xor round xor x"30");
	addcons(311 downto 264) <= input(311 downto 264);  
	addcons(263 downto 256) <= input(263 downto 256);

	addcons(255 downto 248) <= (input(255 downto 248) xor round xor x"40");
	addcons(247 downto 200) <= input(247 downto 200);  
	addcons(199 downto 192) <= input(199 downto 192); 

	addcons(191 downto 184) <= (input(191 downto 184) xor round xor x"50"); 
	addcons(183 downto 136) <= input(183 downto 136);  
	addcons(135 downto 128) <= input(135 downto 128); 

	addcons(127 downto 120) <= (input(127 downto 120) xor round xor x"60"); 
	addcons(119 downto 72) <= input(119 downto 72); 
	addcons(71 downto 64) <= input(71 downto 64);

	addcons(63 downto 56) <= (input(63 downto 56) xor round xor x"70"); 
	addcons(55 downto 8) <= input(55 downto 8);  
	addcons(7 downto 0) <= input(7 downto 0); 
	
	cons256_gen: for i in 0 to GROESTL_DATA_SIZE_SMALL/8-1 generate
		cons_matrix(GROESTL_DATA_SIZE_SMALL/8-i-1) <= addcons(8*i+7 downto 8*i);
	end generate;
	
	to_sbox <= 	cons_matrix(0) & cons_matrix(9) & cons_matrix(18) & cons_matrix(27) &
				cons_matrix(36) & cons_matrix(45) & cons_matrix(54) & cons_matrix(63) &
				cons_matrix(8) & cons_matrix(17) & cons_matrix(26) & cons_matrix(35) &
				cons_matrix(44) & cons_matrix(53) &	cons_matrix(62) & cons_matrix(7) &
				cons_matrix(16) & cons_matrix(25) & cons_matrix(34) & cons_matrix(43) &
				cons_matrix(52) & cons_matrix(61) & cons_matrix(6) & cons_matrix(15) &
				cons_matrix(24) & cons_matrix(33) & cons_matrix(42) & cons_matrix(51) &
				cons_matrix(60) & cons_matrix(5) & cons_matrix(14) & cons_matrix(23) when sel_rd="111" else 
				cons_matrix(32) & cons_matrix(41) & cons_matrix(50) & cons_matrix(59) &
				cons_matrix(4) & cons_matrix(13) & cons_matrix(22) & cons_matrix(31) &
				cons_matrix(40) & cons_matrix(49) & cons_matrix(58) & cons_matrix(3) &
				cons_matrix(12) & cons_matrix(21) &	cons_matrix(30) & cons_matrix(39) &
				cons_matrix(48) & cons_matrix(57) & cons_matrix(2) & cons_matrix(11) &
				cons_matrix(20) & cons_matrix(29) & cons_matrix(38) & cons_matrix(47) &
				cons_matrix(56) & cons_matrix(1) & cons_matrix(10) & cons_matrix(19) &
				cons_matrix(28) & cons_matrix(37) & cons_matrix(46) & cons_matrix(55);				
				
				
end generate;			  		  

gen512: if n=GROESTL_DATA_SIZE_BIG generate	 

	addcons(1023 downto 1016) <= (input(1023 downto 1016) xor round xor x"00");
	addcons(1015 downto 968) <= input(1015 downto 968);
	addcons(967 downto 960) <= input(967 downto 960);

	addcons(959 downto 952) <= (input(959 downto 952) xor round xor x"10");
	addcons(951 downto 904) <= input(951 downto 904);
	addcons(903 downto 896) <= input(903 downto 896);

	addcons(895 downto 888) <= (input(895 downto 888) xor round xor x"20");
	addcons(887 downto 840) <= input(887 downto 840);
	addcons(839 downto 832) <= input(839 downto 832);

	addcons(831 downto 824) <= (input(831 downto 824) xor round xor x"30");
	addcons(823 downto 776) <= input(823 downto 776);
	addcons(775 downto 768) <= input(775 downto 768);

	addcons(767 downto 760) <= (input(767 downto 760) xor round xor x"40");
	addcons(759 downto 712) <= input(759 downto 712);
	addcons(711 downto 704) <= input(711 downto 704);

	addcons(703 downto 696) <= (input(703 downto 696) xor round xor x"50");
	addcons(695 downto 648) <= input(695 downto 648);
	addcons(647 downto 640) <= input(647 downto 640);

	addcons(639 downto 632) <= (input(639 downto 632) xor round xor x"60");
	addcons(631 downto 584) <= input(631 downto 584);
	addcons(583 downto 576) <= input(583 downto 576);

	addcons(575 downto 568) <= (input(575 downto 568) xor round xor x"70");
	addcons(567 downto 520) <= input(567 downto 520);
	addcons(519 downto 512) <= input(519 downto 512);

	addcons(511 downto 504) <= (input(511 downto 504) xor round xor x"80");
	addcons(503 downto 456) <= input(503 downto 456);
	addcons(455 downto 448) <= input(455 downto 448);
	
	addcons(447 downto 440) <= (input(447 downto 440) xor round xor x"90");
	addcons(439 downto 392) <= input(439 downto 392);
	addcons(391 downto 384) <= input(391 downto 384);

	addcons(383 downto 376) <= (input(383 downto 376) xor round xor x"a0");
	addcons(375 downto 328) <= input(375 downto 328);
	addcons(327 downto 320) <= input(327 downto 320);

	addcons(319 downto 312) <= (input(319 downto 312) xor round xor x"b0");
	addcons(311 downto 264) <= input(311 downto 264);
	addcons(263 downto 256) <= input(263 downto 256);

	addcons(255 downto 248) <= (input(255 downto 248) xor round xor x"c0");
	addcons(247 downto 200) <= input(247 downto 200);
	addcons(199 downto 192) <= input(199 downto 192);

	addcons(191 downto 184) <= (input(191 downto 184) xor round xor x"d0");
	addcons(183 downto 136) <= input(183 downto 136);
	addcons(135 downto 128) <= input(135 downto 128);

	addcons(127 downto 120) <= (input(127 downto 120) xor round xor x"e0");
	addcons(119 downto 72) <= input(119 downto 72);
	addcons(71 downto 64) <= input(71 downto 64);

	addcons(63 downto 56) <= (input(63 downto 56) xor round xor x"f0");
	addcons(55 downto 8) <= input(55 downto 8);
	addcons(7 downto 0) <= input(7 downto 0);
	
	cons512_gen: for i in 0 to GROESTL_DATA_SIZE_BIG/8-1 generate
		cons_matrix(GROESTL_DATA_SIZE_BIG/8-i-1) <= addcons(8*i+7 downto 8*i);
	end generate;
	
	to_sbox <= 	cons_matrix(0) & cons_matrix(9) & cons_matrix(18) & cons_matrix(27) &
			cons_matrix(36) & cons_matrix(45) & cons_matrix(54) & cons_matrix(95) &
			cons_matrix(8) & cons_matrix(17) & cons_matrix(26) & cons_matrix(35) &
			cons_matrix(44) & cons_matrix(53) &	cons_matrix(62) & cons_matrix(103) &
			cons_matrix(16) & cons_matrix(25) & cons_matrix(34) & cons_matrix(43) &
			cons_matrix(52) & cons_matrix(61) & cons_matrix(70) & cons_matrix(111) &
			cons_matrix(24) & cons_matrix(33) & cons_matrix(42) & cons_matrix(51) &
			cons_matrix(60) & cons_matrix(69) & cons_matrix(78) & cons_matrix(119) &
			cons_matrix(32) & cons_matrix(41) & cons_matrix(50) & cons_matrix(59) &
			cons_matrix(68) & cons_matrix(77) & cons_matrix(86) & cons_matrix(127) &
			cons_matrix(40) & cons_matrix(49) & cons_matrix(58) & cons_matrix(67) &
			cons_matrix(76) & cons_matrix(85) & cons_matrix(94) & cons_matrix(7) &
			cons_matrix(48) & cons_matrix(57) & cons_matrix(66) & cons_matrix(75) &
			cons_matrix(84) & cons_matrix(93) & cons_matrix(102) & cons_matrix(15) &
			cons_matrix(56) & cons_matrix(65) & cons_matrix(74) & cons_matrix(83) &			
			cons_matrix(92) & cons_matrix(101) & cons_matrix(110) & cons_matrix(23) when sel_rd="111" else 
			cons_matrix(64) & cons_matrix(73) & cons_matrix(82) & cons_matrix(91) &
			cons_matrix(100) & cons_matrix(109) & cons_matrix(118) & cons_matrix(31) &
			cons_matrix(72) & cons_matrix(81) & cons_matrix(90) & cons_matrix(99) &
			cons_matrix(108) & cons_matrix(117) &	cons_matrix(126) & cons_matrix(39) &
			cons_matrix(80) & cons_matrix(89) & cons_matrix(98) & cons_matrix(107) &
			cons_matrix(116) & cons_matrix(125) & cons_matrix(6) & cons_matrix(47) &
			cons_matrix(88) & cons_matrix(97) & cons_matrix(106) & cons_matrix(115) &
			cons_matrix(124) & cons_matrix(5) & cons_matrix(14) & cons_matrix(55) &
			cons_matrix(96) & cons_matrix(105) & cons_matrix(114) & cons_matrix(123) &
			cons_matrix(4) & cons_matrix(13) & cons_matrix(22) & cons_matrix(63) &
			cons_matrix(104) & cons_matrix(113) & cons_matrix(122) & cons_matrix(3) &
			cons_matrix(12) & cons_matrix(21) & cons_matrix(30) & cons_matrix(71) &
			cons_matrix(112) & cons_matrix(121) & cons_matrix(2) & cons_matrix(11) &
			cons_matrix(20) & cons_matrix(29) & cons_matrix(38) & cons_matrix(79) &
			cons_matrix(120) & cons_matrix(1) & cons_matrix(10) & cons_matrix(19) &			
			cons_matrix(28) & cons_matrix(37) & cons_matrix(46) & cons_matrix(87);	
			 
end generate;
		
	sbox_gen: for i in 0 to n/(2*AES_SBOX_SIZE) - 1  generate
	sbox	: aes_sbox 	generic map (rom_style=>rom_style)
			port map (	 
				input=>to_sbox(AES_SBOX_SIZE*i + 7 downto AES_SBOX_SIZE*i), 
				output=>after_subbyte(AES_SBOX_SIZE*i+7 downto AES_SBOX_SIZE*i));	
	end generate;	
	
	after_shiftrow <= after_subbyte;

mc256: if n=GROESTL_DATA_SIZE_SMALL generate	
	mc			: entity work.groestl_mixbytes(groestl_mixbytes)
				generic map (n=>n/2)
				port map (input=>after_shiftrow,  output=>after_mixcolumn);	
end generate;

mc512: if n=GROESTL_DATA_SIZE_BIG generate	
	mc_left			: entity work.groestl_mixbytes(groestl_mixbytes)
					generic map (n=>n/2)	
					port map (input=>after_shiftrow,  output=>after_mixcolumn);	 
					
						
end generate;

	reg		: regn generic map (n=>n/2, init=>zero) port map (clk=>clk, rst=>rst, en=>VCC, input=>after_mixcolumn, output=>after_reg);

	output <= after_reg & after_mixcolumn;	
end p_folded_fx2; 


architecture q_folded_fx2 of groestl_pq_parallel_folded is

signal	after_subbyte		: std_logic_vector(n/2-1 downto 0);
signal	addcons			: std_logic_vector(n-1 downto 0);
signal	after_reg		: std_logic_vector(n/2-1 downto 0);
signal	after_shiftrow	: std_logic_vector(n/2-1 downto 0);
constant zero			: std_logic_vector(n/2-1 downto 0):=(others=>'0'); 
signal after_mixcolumn, to_sbox	 : std_logic_vector(n/2-1 downto 0);
type matrix is array (0 to n/8-1) of std_logic_vector(7 downto 0);
signal cons_matrix	: matrix;
begin
	
	
gen256: if n=GROESTL_DATA_SIZE_SMALL generate

	addcons(511 downto 504) <=  (not input(511 downto 504));
	addcons(503 downto 456) <=  (not input(503 downto 456)); 
	addcons(455 downto 448) <=  (x"ff" xor input(455 downto 448) xor round);
	
	addcons(447 downto 440) <=  (not input(447 downto 440));
	addcons(439 downto 392) <=  (not input(439 downto 392)); 
	addcons(391 downto 384) <=  (x"ef" xor input(391 downto 384) xor round);

	addcons(383 downto 376) <=  (not input(383 downto 376));
	addcons(375 downto 328) <= (not input(375 downto 328)); 
	addcons(327 downto 320) <= (x"df" xor input(327 downto 320) xor round);

	addcons(319 downto 312) <=  (not input(319 downto 312));
	addcons(311 downto 264) <= (not input(311 downto 264)); 
	addcons(263 downto 256) <= (x"cf" xor input(263 downto 256) xor round);

	addcons(255 downto 248) <=  (not input(255 downto 248));
	addcons(247 downto 200) <=  (not input(247 downto 200)); 
	addcons(199 downto 192) <=  (x"bf" xor input(199 downto 192) xor round);

	addcons(191 downto 184) <=  (not input(191 downto 184));
	addcons(183 downto 136) <=  (not input(183 downto 136)); 
	addcons(135 downto 128) <=  (x"af" xor input(135 downto 128) xor round);

	addcons(127 downto 120) <=  (not input(127 downto 120));
	addcons(119 downto 72) <=  (not input(119 downto 72)); 
	addcons(71 downto 64) <=  (x"9f" xor input(71 downto 64) xor round);

	addcons(63 downto 56) <=  (not input(63 downto 56));
	addcons(55 downto 8) <=  (not input(55 downto 8)); 
	addcons(7 downto 0) <=  (x"8f" xor input(7 downto 0) xor round);
		
	cons256_gen: for i in 0 to GROESTL_DATA_SIZE_SMALL/8-1 generate
		cons_matrix(GROESTL_DATA_SIZE_SMALL/8-i-1) <= addcons(8*i+7 downto 8*i);
	end generate;
	
		to_sbox <= 	cons_matrix(8) & cons_matrix(25) & cons_matrix(42) & cons_matrix(59) &
				cons_matrix(4) & cons_matrix(21) & cons_matrix(38) & cons_matrix(55) &
				cons_matrix(16) & cons_matrix(33) & cons_matrix(50) & cons_matrix(3) &
				cons_matrix(12) & cons_matrix(29) &	cons_matrix(46) & cons_matrix(63) &
				cons_matrix(24) & cons_matrix(41) & cons_matrix(58) & cons_matrix(11) &
				cons_matrix(20) & cons_matrix(37) & cons_matrix(54) & cons_matrix(7) &
				cons_matrix(32) & cons_matrix(49) & cons_matrix(2) & cons_matrix(19) &
				cons_matrix(28) & cons_matrix(45) & cons_matrix(62) & cons_matrix(15) when sel_rd="111" else 
				cons_matrix(40) & cons_matrix(57) & cons_matrix(10) & cons_matrix(27) &
				cons_matrix(36) & cons_matrix(53) & cons_matrix(6) & cons_matrix(23) &
				cons_matrix(48) & cons_matrix(1) & cons_matrix(18) & cons_matrix(35) &
				cons_matrix(44) & cons_matrix(61) &	cons_matrix(14) & cons_matrix(31) &
				cons_matrix(56) & cons_matrix(9) & cons_matrix(26) & cons_matrix(43) &
				cons_matrix(52) & cons_matrix(5) & cons_matrix(22) & cons_matrix(39) &
				cons_matrix(0) & cons_matrix(17) & cons_matrix(34) & cons_matrix(51) &
				cons_matrix(60) & cons_matrix(13) & cons_matrix(30) & cons_matrix(47);	
	
end generate;			  		  

gen512: if n=GROESTL_DATA_SIZE_BIG generate	 

	addcons(1023 downto 1016) <=  (not input(1023 downto 1016));
	addcons(1015 downto 968) <=  (not input(1015 downto 968)); 
	addcons(967 downto 960) <=  (x"ff" xor input(967 downto 960) xor round); 

	addcons(959 downto 952) <=  (not input(959 downto 952));
	addcons(951 downto 904) <=  (not input(951 downto 904)); 
	addcons(903 downto 896) <=  (x"ef" xor input(903 downto 896) xor round); 

	addcons(895 downto 888) <=  (not input(895 downto 888));
	addcons(887 downto 840) <=  (not input(887 downto 840)); 
	addcons(839 downto 832) <=  (x"df" xor input(839 downto 832) xor round);  

	addcons(831 downto 824) <=  (not input(831 downto 824));
	addcons(823 downto 776) <=  (not input(823 downto 776)); 
	addcons(775 downto 768) <=  (x"cf" xor input(775 downto 768) xor round);

	addcons(767 downto 760) <=  (not input(767 downto 760));
	addcons(759 downto 712) <=  (not input(759 downto 712)); 
	addcons(711 downto 704) <=  (x"bf" xor input(711 downto 704) xor round);

	addcons(703 downto 696) <=  (not input(703 downto 696));
	addcons(695 downto 648) <=  (not input(695 downto 648)); 
	addcons(647 downto 640) <=  (x"af" xor input(647 downto 640) xor round);

	addcons(639 downto 632) <=  (not input(639 downto 632));
	addcons(631 downto 584) <=  (not input(631 downto 584)); 
	addcons(583 downto 576) <=  (x"9f" xor input(583 downto 576) xor round);

	addcons(575 downto 568) <=  (not input(575 downto 568));
	addcons(567 downto 520) <= (not input(567 downto 520)); 
	addcons(519 downto 512) <=  (x"8f" xor input(519 downto 512) xor round);

	addcons(511 downto 504) <=  (not input(511 downto 504));
	addcons(503 downto 456) <=  (not input(503 downto 456)); 
	addcons(455 downto 448) <=  (x"7f" xor input(455 downto 448) xor round);
	
	addcons(447 downto 440) <=  (not input(447 downto 440));
	addcons(439 downto 392) <=  (not input(439 downto 392)); 
	addcons(391 downto 384) <=  (x"6f" xor input(391 downto 384) xor round);

	addcons(383 downto 376) <=  (not input(383 downto 376));
	addcons(375 downto 328) <=  (not input(375 downto 328)); 
	addcons(327 downto 320) <=  (x"5f" xor input(327 downto 320) xor round);

	addcons(319 downto 312) <=  (not input(319 downto 312));
	addcons(311 downto 264) <=  (not input(311 downto 264)); 
	addcons(263 downto 256) <=  (x"4f" xor input(263 downto 256) xor round);

	addcons(255 downto 248) <=  (not input(255 downto 248));
	addcons(247 downto 200) <=  (not input(247 downto 200)); 
	addcons(199 downto 192) <=  (x"3f" xor input(199 downto 192) xor round);

	addcons(191 downto 184) <=  (not input(191 downto 184));
	addcons(183 downto 136) <=  (not input(183 downto 136)); 
	addcons(135 downto 128) <=  (x"2f" xor input(135 downto 128) xor round);

	addcons(127 downto 120) <=  (not input(127 downto 120));
	addcons(119 downto 72) <=  (not input(119 downto 72)); 
	addcons(71 downto 64) <=  (x"1f" xor input(71 downto 64) xor round);

	addcons(63 downto 56) <=  (not input(63 downto 56));
	addcons(55 downto 8) <=  (not input(55 downto 8)); 
	addcons(7 downto 0) <=  (x"0f" xor input(7 downto 0) xor round);

	cons512_gen: for i in 0 to GROESTL_DATA_SIZE_BIG/8-1 generate
		cons_matrix(GROESTL_DATA_SIZE_BIG/8-i-1) <= addcons(8*i+7 downto 8*i);
	end generate;
	
	to_sbox <= 	cons_matrix(8) & cons_matrix(25) & cons_matrix(42) & cons_matrix(91) &
			cons_matrix(4) & cons_matrix(21) & cons_matrix(38) & cons_matrix(55) &
			cons_matrix(16) & cons_matrix(33) & cons_matrix(50) & cons_matrix(99) &
			cons_matrix(12) & cons_matrix(29) &	cons_matrix(46) & cons_matrix(63) &
			cons_matrix(24) & cons_matrix(41) & cons_matrix(58) & cons_matrix(107) &
			cons_matrix(20) & cons_matrix(37) & cons_matrix(54) & cons_matrix(71) &
			cons_matrix(32) & cons_matrix(49) & cons_matrix(66) & cons_matrix(115) &
			cons_matrix(28) & cons_matrix(45) & cons_matrix(62) & cons_matrix(79) &
			cons_matrix(40) & cons_matrix(57) & cons_matrix(74) & cons_matrix(123) &
			cons_matrix(36) & cons_matrix(53) & cons_matrix(70) & cons_matrix(87) &
			cons_matrix(48) & cons_matrix(65) & cons_matrix(82) & cons_matrix(3) &
			cons_matrix(44) & cons_matrix(61) & cons_matrix(78) & cons_matrix(95) &
			cons_matrix(56) & cons_matrix(73) & cons_matrix(90) & cons_matrix(11) &
			cons_matrix(52) & cons_matrix(69) & cons_matrix(86) & cons_matrix(103) &
			cons_matrix(64) & cons_matrix(81) & cons_matrix(98) & cons_matrix(19) &			
			cons_matrix(60) & cons_matrix(77) & cons_matrix(94) & cons_matrix(111) when sel_rd="111" else 
			cons_matrix(72) & cons_matrix(89) & cons_matrix(106) & cons_matrix(27) &
			cons_matrix(68) & cons_matrix(85) & cons_matrix(102) & cons_matrix(119) &
			cons_matrix(80) & cons_matrix(97) & cons_matrix(114) & cons_matrix(35) &
			cons_matrix(76) & cons_matrix(93) & cons_matrix(110) & cons_matrix(127) &
			cons_matrix(88) & cons_matrix(105) & cons_matrix(122) & cons_matrix(43) &
			cons_matrix(84) & cons_matrix(101) & cons_matrix(118) & cons_matrix(7) &
			cons_matrix(96) & cons_matrix(113) & cons_matrix(2) & cons_matrix(51) &
			cons_matrix(92) & cons_matrix(109) & cons_matrix(126) & cons_matrix(15) &
			cons_matrix(104) & cons_matrix(121) & cons_matrix(10) & cons_matrix(59) &
			cons_matrix(100) & cons_matrix(117) & cons_matrix(6) & cons_matrix(23) &
			cons_matrix(112) & cons_matrix(1) & cons_matrix(18) & cons_matrix(67) &
			cons_matrix(108) & cons_matrix(125) & cons_matrix(14) & cons_matrix(31) &
			cons_matrix(120) & cons_matrix(9) & cons_matrix(26) & cons_matrix(75) &
			cons_matrix(116) & cons_matrix(5) & cons_matrix(22) & cons_matrix(39) &
			cons_matrix(0) & cons_matrix(17) & cons_matrix(34) & cons_matrix(83) &			
			cons_matrix(124) & cons_matrix(13) & cons_matrix(30) & cons_matrix(47);
	
end generate;
	
	
	sbox_gen: for i in 0 to n/(2*AES_SBOX_SIZE) - 1  generate
	sbox	: aes_sbox 	generic map (rom_style=>rom_style)
			port map (	 
				input=>to_sbox(AES_SBOX_SIZE*i + 7 downto AES_SBOX_SIZE*i), 
				output=>after_subbyte(AES_SBOX_SIZE*i+7 downto AES_SBOX_SIZE*i));	
	end generate;	

	after_shiftrow <= after_subbyte;

mc256: if n=GROESTL_DATA_SIZE_SMALL generate	
	mc			: entity work.groestl_mixbytes(groestl_mixbytes)	
				generic map (n=>n/2)
				port map (input=>after_shiftrow,  output=>after_mixcolumn);	
end generate;

mc512: if n=GROESTL_DATA_SIZE_BIG generate	
	mc_left			: entity work.groestl_mixbytes(groestl_mixbytes)	
					generic map (n=>n/2)
					port map (input=>after_shiftrow,  output=>after_mixcolumn);	

end generate;

	reg		: regn generic map (n=>n/2, init=>zero) port map (clk=>clk, rst=>rst, en=>VCC, input=>after_mixcolumn, output=>after_reg);

	output <= after_reg & after_mixcolumn;
	
end q_folded_fx2; 



architecture p_folded_fx4 of groestl_pq_parallel_folded is

signal	after_subbyte		: std_logic_vector(n/4-1 downto 0);
signal	addcons			: std_logic_vector(n-1 downto 0);
signal	after_rega,after_regb,after_regc 		: std_logic_vector(n/4-1 downto 0);
signal to_sbox			: std_logic_vector(n/4-1 downto 0);
signal	after_shiftrow	: std_logic_vector(n/4-1 downto 0);
constant zero			: std_logic_vector(n/4-1 downto 0):=(others=>'0');
signal after_mixcolumn	 : std_logic_vector(n/4-1 downto 0);
type matrix is array (0 to n/8) of std_logic_vector(7 downto 0);
signal cons_matrix	: matrix;

begin
	
	
gen256: if n=GROESTL_DATA_SIZE_SMALL generate

	addcons(511 downto 504) <= (input(511 downto 504) xor round xor x"00"); 
	addcons(503 downto 456) <= input(503 downto 456); 
	addcons(455 downto 448) <= input(455 downto 448); 
	
	addcons(447 downto 440) <= (input(447 downto 440) xor round xor x"10");
	addcons(439 downto 392) <= input(439 downto 392); 
	addcons(391 downto 384) <= input(391 downto 384); 

	addcons(383 downto 376) <= (input(383 downto 376) xor round xor x"20"); 
	addcons(375 downto 328) <= input(375 downto 328);  
	addcons(327 downto 320) <= input(327 downto 320); 

	addcons(319 downto 312) <= (input(319 downto 312) xor round xor x"30");
	addcons(311 downto 264) <= input(311 downto 264);  
	addcons(263 downto 256) <= input(263 downto 256);

	addcons(255 downto 248) <= (input(255 downto 248) xor round xor x"40");
	addcons(247 downto 200) <= input(247 downto 200);  
	addcons(199 downto 192) <= input(199 downto 192); 

	addcons(191 downto 184) <= (input(191 downto 184) xor round xor x"50"); 
	addcons(183 downto 136) <= input(183 downto 136);  
	addcons(135 downto 128) <= input(135 downto 128); 

	addcons(127 downto 120) <= (input(127 downto 120) xor round xor x"60"); 
	addcons(119 downto 72) <= input(119 downto 72); 
	addcons(71 downto 64) <= input(71 downto 64);

	addcons(63 downto 56) <= (input(63 downto 56) xor round xor x"70"); 
	addcons(55 downto 8) <= input(55 downto 8);  
	addcons(7 downto 0) <= input(7 downto 0); 
	
	cons256_gen: for i in 0 to GROESTL_DATA_SIZE_SMALL/8-1 generate
		cons_matrix(GROESTL_DATA_SIZE_SMALL/8-i-1) <= addcons(8*i+7 downto 8*i);
	end generate;
	
	to_sbox <= 	cons_matrix(0) & cons_matrix(9) & cons_matrix(18) & cons_matrix(27) &
				cons_matrix(36) & cons_matrix(45) & cons_matrix(54) & cons_matrix(63) &
				cons_matrix(8) & cons_matrix(17) & cons_matrix(26) & cons_matrix(35) &
				cons_matrix(44) & cons_matrix(53) &	cons_matrix(62) & cons_matrix(7) when sel_rd="001" else
				cons_matrix(16) & cons_matrix(25) & cons_matrix(34) & cons_matrix(43) &
				cons_matrix(52) & cons_matrix(61) & cons_matrix(6) & cons_matrix(15) &
				cons_matrix(24) & cons_matrix(33) & cons_matrix(42) & cons_matrix(51) &
				cons_matrix(60) & cons_matrix(5) & cons_matrix(14) & cons_matrix(23) when sel_rd="010" else 
				cons_matrix(32) & cons_matrix(41) & cons_matrix(50) & cons_matrix(59) &
				cons_matrix(4) & cons_matrix(13) & cons_matrix(22) & cons_matrix(31) &
				cons_matrix(40) & cons_matrix(49) & cons_matrix(58) & cons_matrix(3) &
				cons_matrix(12) & cons_matrix(21) &	cons_matrix(30) & cons_matrix(39) when sel_rd="100" else
				cons_matrix(48) & cons_matrix(57) & cons_matrix(2) & cons_matrix(11) &
				cons_matrix(20) & cons_matrix(29) & cons_matrix(38) & cons_matrix(47) &
				cons_matrix(56) & cons_matrix(1) & cons_matrix(10) & cons_matrix(19) &
				cons_matrix(28) & cons_matrix(37) & cons_matrix(46) & cons_matrix(55);				
				
				
end generate;			  		  

gen512: if n=GROESTL_DATA_SIZE_BIG generate	 

	addcons(1023 downto 1016) <= (input(1023 downto 1016) xor round xor x"00");
	addcons(1015 downto 968) <= input(1015 downto 968);
	addcons(967 downto 960) <= input(967 downto 960);

	addcons(959 downto 952) <= (input(959 downto 952) xor round xor x"10");
	addcons(951 downto 904) <= input(951 downto 904);
	addcons(903 downto 896) <= input(903 downto 896);

	addcons(895 downto 888) <= (input(895 downto 888) xor round xor x"20");
	addcons(887 downto 840) <= input(887 downto 840);
	addcons(839 downto 832) <= input(839 downto 832);

	addcons(831 downto 824) <= (input(831 downto 824) xor round xor x"30");
	addcons(823 downto 776) <= input(823 downto 776);
	addcons(775 downto 768) <= input(775 downto 768);

	addcons(767 downto 760) <= (input(767 downto 760) xor round xor x"40");
	addcons(759 downto 712) <= input(759 downto 712);
	addcons(711 downto 704) <= input(711 downto 704);

	addcons(703 downto 696) <= (input(703 downto 696) xor round xor x"50");
	addcons(695 downto 648) <= input(695 downto 648);
	addcons(647 downto 640) <= input(647 downto 640);

	addcons(639 downto 632) <= (input(639 downto 632) xor round xor x"60");
	addcons(631 downto 584) <= input(631 downto 584);
	addcons(583 downto 576) <= input(583 downto 576);

	addcons(575 downto 568) <= (input(575 downto 568) xor round xor x"70");
	addcons(567 downto 520) <= input(567 downto 520);
	addcons(519 downto 512) <= input(519 downto 512);

	addcons(511 downto 504) <= (input(511 downto 504) xor round xor x"80");
	addcons(503 downto 456) <= input(503 downto 456);
	addcons(455 downto 448) <= input(455 downto 448);
	
	addcons(447 downto 440) <= (input(447 downto 440) xor round xor x"90");
	addcons(439 downto 392) <= input(439 downto 392);
	addcons(391 downto 384) <= input(391 downto 384);

	addcons(383 downto 376) <= (input(383 downto 376) xor round xor x"a0");
	addcons(375 downto 328) <= input(375 downto 328);
	addcons(327 downto 320) <= input(327 downto 320);

	addcons(319 downto 312) <= (input(319 downto 312) xor round xor x"b0");
	addcons(311 downto 264) <= input(311 downto 264);
	addcons(263 downto 256) <= input(263 downto 256);

	addcons(255 downto 248) <= (input(255 downto 248) xor round xor x"c0");
	addcons(247 downto 200) <= input(247 downto 200);
	addcons(199 downto 192) <= input(199 downto 192);

	addcons(191 downto 184) <= (input(191 downto 184) xor round xor x"d0");
	addcons(183 downto 136) <= input(183 downto 136);
	addcons(135 downto 128) <= input(135 downto 128);

	addcons(127 downto 120) <= (input(127 downto 120) xor round xor x"e0");
	addcons(119 downto 72) <= input(119 downto 72);
	addcons(71 downto 64) <= input(71 downto 64);

	addcons(63 downto 56) <= (input(63 downto 56) xor round xor x"f0");
	addcons(55 downto 8) <= input(55 downto 8);
	addcons(7 downto 0) <= input(7 downto 0);
	
	cons512_gen: for i in 0 to GROESTL_DATA_SIZE_BIG/8-1 generate
		cons_matrix(GROESTL_DATA_SIZE_BIG/8-i-1) <= addcons(8*i+7 downto 8*i);
	end generate;
	
	to_sbox <= 	cons_matrix(0) & cons_matrix(9) & cons_matrix(18) & cons_matrix(27) &
			cons_matrix(36) & cons_matrix(45) & cons_matrix(54) & cons_matrix(95) &
			cons_matrix(8) & cons_matrix(17) & cons_matrix(26) & cons_matrix(35) &
			cons_matrix(44) & cons_matrix(53) &	cons_matrix(62) & cons_matrix(103) &
			cons_matrix(16) & cons_matrix(25) & cons_matrix(34) & cons_matrix(43) &
			cons_matrix(52) & cons_matrix(61) & cons_matrix(70) & cons_matrix(111) &
			cons_matrix(24) & cons_matrix(33) & cons_matrix(42) & cons_matrix(51) &
			cons_matrix(60) & cons_matrix(69) & cons_matrix(78) & cons_matrix(119) when sel_rd="001" else
			cons_matrix(32) & cons_matrix(41) & cons_matrix(50) & cons_matrix(59) &
			cons_matrix(68) & cons_matrix(77) & cons_matrix(86) & cons_matrix(127) &
			cons_matrix(40) & cons_matrix(49) & cons_matrix(58) & cons_matrix(67) &
			cons_matrix(76) & cons_matrix(85) & cons_matrix(94) & cons_matrix(7) &
			cons_matrix(48) & cons_matrix(57) & cons_matrix(66) & cons_matrix(75) &
			cons_matrix(84) & cons_matrix(93) & cons_matrix(102) & cons_matrix(15) &
			cons_matrix(56) & cons_matrix(65) & cons_matrix(74) & cons_matrix(83) &			
			cons_matrix(92) & cons_matrix(101) & cons_matrix(110) & cons_matrix(23) when sel_rd="010" else 
			cons_matrix(64) & cons_matrix(73) & cons_matrix(82) & cons_matrix(91) &
			cons_matrix(100) & cons_matrix(109) & cons_matrix(118) & cons_matrix(31) &
			cons_matrix(72) & cons_matrix(81) & cons_matrix(90) & cons_matrix(99) &
			cons_matrix(108) & cons_matrix(117) &	cons_matrix(126) & cons_matrix(39) &
			cons_matrix(80) & cons_matrix(89) & cons_matrix(98) & cons_matrix(107) &
			cons_matrix(116) & cons_matrix(125) & cons_matrix(6) & cons_matrix(47) &
			cons_matrix(88) & cons_matrix(97) & cons_matrix(106) & cons_matrix(115) &
			cons_matrix(124) & cons_matrix(5) & cons_matrix(14) & cons_matrix(55) when sel_rd="100" else
			cons_matrix(96) & cons_matrix(105) & cons_matrix(114) & cons_matrix(123) &
			cons_matrix(4) & cons_matrix(13) & cons_matrix(22) & cons_matrix(63) &
			cons_matrix(104) & cons_matrix(113) & cons_matrix(122) & cons_matrix(3) &
			cons_matrix(12) & cons_matrix(21) & cons_matrix(30) & cons_matrix(71) &
			cons_matrix(112) & cons_matrix(121) & cons_matrix(2) & cons_matrix(11) &
			cons_matrix(20) & cons_matrix(29) & cons_matrix(38) & cons_matrix(79) &
			cons_matrix(120) & cons_matrix(1) & cons_matrix(10) & cons_matrix(19) &			
			cons_matrix(28) & cons_matrix(37) & cons_matrix(46) & cons_matrix(87);	
			 
end generate;
		
	sbox_gen: for i in 0 to n/(4*AES_SBOX_SIZE) - 1  generate
	sbox	: aes_sbox 	generic map (rom_style=>rom_style)
			port map (	 
				input=>to_sbox(AES_SBOX_SIZE*i + 7 downto AES_SBOX_SIZE*i), 
				output=>after_subbyte(AES_SBOX_SIZE*i+7 downto AES_SBOX_SIZE*i));	
	end generate;	
	
	after_shiftrow <= after_subbyte;

mc256: if n=GROESTL_DATA_SIZE_SMALL generate	
	mc			: entity work.groestl_mixbytes(groestl_mixbytes)
				generic map (n=>n/4)
				port map (input=>after_shiftrow,  output=>after_mixcolumn);	
end generate;

mc512: if n=GROESTL_DATA_SIZE_BIG generate	
	mc_left			: entity work.groestl_mixbytes(groestl_mixbytes)
					generic map (n=>n/4)	
					port map (input=>after_shiftrow,  output=>after_mixcolumn);	
end generate;

	reg00		: regn generic map (n=>n/4, init=>zero) port map (clk=>clk, rst=>rst, en=>VCC, input=>after_mixcolumn, output=>after_rega);
	reg01		: regn generic map (n=>n/4, init=>zero) port map (clk=>clk, rst=>rst, en=>VCC, input=>after_rega, output=>after_regb);
	reg02		: regn generic map (n=>n/4, init=>zero) port map (clk=>clk, rst=>rst, en=>VCC, input=>after_regb, output=>after_regc);

	output <= after_regc & after_regb & after_rega & after_mixcolumn;	
end p_folded_fx4; 


architecture q_folded_fx4 of groestl_pq_parallel_folded is

signal	after_subbyte		: std_logic_vector(n/4-1 downto 0);
signal	addcons			: std_logic_vector(n-1 downto 0);
signal	after_rega,after_regb,after_regc 		: std_logic_vector(n/4-1 downto 0);
signal	after_shiftrow	: std_logic_vector(n/4-1 downto 0);
constant zero			: std_logic_vector(n/4-1 downto 0):=(others=>'0'); 
signal after_mixcolumn, to_sbox	 : std_logic_vector(n/4-1 downto 0);
type matrix is array (0 to n/8-1) of std_logic_vector(7 downto 0);
signal cons_matrix	: matrix;
begin
	
	
gen256: if n=GROESTL_DATA_SIZE_SMALL generate

	addcons(511 downto 504) <=  (not input(511 downto 504));
	addcons(503 downto 456) <=  (not input(503 downto 456)); 
	addcons(455 downto 448) <=  (x"ff" xor input(455 downto 448) xor round);
	
	addcons(447 downto 440) <=  (not input(447 downto 440));
	addcons(439 downto 392) <=  (not input(439 downto 392)); 
	addcons(391 downto 384) <=  (x"ef" xor input(391 downto 384) xor round);

	addcons(383 downto 376) <=  (not input(383 downto 376));
	addcons(375 downto 328) <= (not input(375 downto 328)); 
	addcons(327 downto 320) <= (x"df" xor input(327 downto 320) xor round);

	addcons(319 downto 312) <=  (not input(319 downto 312));
	addcons(311 downto 264) <= (not input(311 downto 264)); 
	addcons(263 downto 256) <= (x"cf" xor input(263 downto 256) xor round);

	addcons(255 downto 248) <=  (not input(255 downto 248));
	addcons(247 downto 200) <=  (not input(247 downto 200)); 
	addcons(199 downto 192) <=  (x"bf" xor input(199 downto 192) xor round);

	addcons(191 downto 184) <=  (not input(191 downto 184));
	addcons(183 downto 136) <=  (not input(183 downto 136)); 
	addcons(135 downto 128) <=  (x"af" xor input(135 downto 128) xor round);

	addcons(127 downto 120) <=  (not input(127 downto 120));
	addcons(119 downto 72) <=  (not input(119 downto 72)); 
	addcons(71 downto 64) <=  (x"9f" xor input(71 downto 64) xor round);

	addcons(63 downto 56) <=  (not input(63 downto 56));
	addcons(55 downto 8) <=  (not input(55 downto 8)); 
	addcons(7 downto 0) <=  (x"8f" xor input(7 downto 0) xor round);
		
	cons256_gen: for i in 0 to GROESTL_DATA_SIZE_SMALL/8-1 generate
		cons_matrix(GROESTL_DATA_SIZE_SMALL/8-i-1) <= addcons(8*i+7 downto 8*i);
	end generate;
	
		to_sbox <= 	cons_matrix(8) & cons_matrix(25) & cons_matrix(42) & cons_matrix(59) &
				cons_matrix(4) & cons_matrix(21) & cons_matrix(38) & cons_matrix(55) &
				cons_matrix(16) & cons_matrix(33) & cons_matrix(50) & cons_matrix(3) &
				cons_matrix(12) & cons_matrix(29) &	cons_matrix(46) & cons_matrix(63) when sel_rd="001" else
				cons_matrix(24) & cons_matrix(41) & cons_matrix(58) & cons_matrix(11) &
				cons_matrix(20) & cons_matrix(37) & cons_matrix(54) & cons_matrix(7) &
				cons_matrix(32) & cons_matrix(49) & cons_matrix(2) & cons_matrix(19) &
				cons_matrix(28) & cons_matrix(45) & cons_matrix(62) & cons_matrix(15) when sel_rd="010" else 
				cons_matrix(40) & cons_matrix(57) & cons_matrix(10) & cons_matrix(27) &
				cons_matrix(36) & cons_matrix(53) & cons_matrix(6) & cons_matrix(23) &
				cons_matrix(48) & cons_matrix(1) & cons_matrix(18) & cons_matrix(35) &
				cons_matrix(44) & cons_matrix(61) &	cons_matrix(14) & cons_matrix(31) when sel_rd="100" else
				cons_matrix(56) & cons_matrix(9) & cons_matrix(26) & cons_matrix(43) &
				cons_matrix(52) & cons_matrix(5) & cons_matrix(22) & cons_matrix(39) &
				cons_matrix(0) & cons_matrix(17) & cons_matrix(34) & cons_matrix(51) &
				cons_matrix(60) & cons_matrix(13) & cons_matrix(30) & cons_matrix(47);	
	
end generate;			  		  

gen512: if n=GROESTL_DATA_SIZE_BIG generate	 

	addcons(1023 downto 1016) <=  (not input(1023 downto 1016));
	addcons(1015 downto 968) <=  (not input(1015 downto 968)); 
	addcons(967 downto 960) <=  (x"ff" xor input(967 downto 960) xor round); 

	addcons(959 downto 952) <=  (not input(959 downto 952));
	addcons(951 downto 904) <=  (not input(951 downto 904)); 
	addcons(903 downto 896) <=  (x"ef" xor input(903 downto 896) xor round); 

	addcons(895 downto 888) <=  (not input(895 downto 888));
	addcons(887 downto 840) <=  (not input(887 downto 840)); 
	addcons(839 downto 832) <=  (x"df" xor input(839 downto 832) xor round);  

	addcons(831 downto 824) <=  (not input(831 downto 824));
	addcons(823 downto 776) <=  (not input(823 downto 776)); 
	addcons(775 downto 768) <=  (x"cf" xor input(775 downto 768) xor round);

	addcons(767 downto 760) <=  (not input(767 downto 760));
	addcons(759 downto 712) <=  (not input(759 downto 712)); 
	addcons(711 downto 704) <=  (x"bf" xor input(711 downto 704) xor round);

	addcons(703 downto 696) <=  (not input(703 downto 696));
	addcons(695 downto 648) <=  (not input(695 downto 648)); 
	addcons(647 downto 640) <=  (x"af" xor input(647 downto 640) xor round);

	addcons(639 downto 632) <=  (not input(639 downto 632));
	addcons(631 downto 584) <=  (not input(631 downto 584)); 
	addcons(583 downto 576) <=  (x"9f" xor input(583 downto 576) xor round);

	addcons(575 downto 568) <=  (not input(575 downto 568));
	addcons(567 downto 520) <= (not input(567 downto 520)); 
	addcons(519 downto 512) <=  (x"8f" xor input(519 downto 512) xor round);

	addcons(511 downto 504) <=  (not input(511 downto 504));
	addcons(503 downto 456) <=  (not input(503 downto 456)); 
	addcons(455 downto 448) <=  (x"7f" xor input(455 downto 448) xor round);
	
	addcons(447 downto 440) <=  (not input(447 downto 440));
	addcons(439 downto 392) <=  (not input(439 downto 392)); 
	addcons(391 downto 384) <=  (x"6f" xor input(391 downto 384) xor round);

	addcons(383 downto 376) <=  (not input(383 downto 376));
	addcons(375 downto 328) <=  (not input(375 downto 328)); 
	addcons(327 downto 320) <=  (x"5f" xor input(327 downto 320) xor round);

	addcons(319 downto 312) <=  (not input(319 downto 312));
	addcons(311 downto 264) <=  (not input(311 downto 264)); 
	addcons(263 downto 256) <=  (x"4f" xor input(263 downto 256) xor round);

	addcons(255 downto 248) <=  (not input(255 downto 248));
	addcons(247 downto 200) <=  (not input(247 downto 200)); 
	addcons(199 downto 192) <=  (x"3f" xor input(199 downto 192) xor round);

	addcons(191 downto 184) <=  (not input(191 downto 184));
	addcons(183 downto 136) <=  (not input(183 downto 136)); 
	addcons(135 downto 128) <=  (x"2f" xor input(135 downto 128) xor round);

	addcons(127 downto 120) <=  (not input(127 downto 120));
	addcons(119 downto 72) <=  (not input(119 downto 72)); 
	addcons(71 downto 64) <=  (x"1f" xor input(71 downto 64) xor round);

	addcons(63 downto 56) <=  (not input(63 downto 56));
	addcons(55 downto 8) <=  (not input(55 downto 8)); 
	addcons(7 downto 0) <=  (x"0f" xor input(7 downto 0) xor round);

	cons512_gen: for i in 0 to GROESTL_DATA_SIZE_BIG/8-1 generate
		cons_matrix(GROESTL_DATA_SIZE_BIG/8-i-1) <= addcons(8*i+7 downto 8*i);
	end generate;
	
	to_sbox <= 	cons_matrix(8) & cons_matrix(25) & cons_matrix(42) & cons_matrix(91) &
			cons_matrix(4) & cons_matrix(21) & cons_matrix(38) & cons_matrix(55) &
			cons_matrix(16) & cons_matrix(33) & cons_matrix(50) & cons_matrix(99) &
			cons_matrix(12) & cons_matrix(29) &	cons_matrix(46) & cons_matrix(63) &
			cons_matrix(24) & cons_matrix(41) & cons_matrix(58) & cons_matrix(107) &
			cons_matrix(20) & cons_matrix(37) & cons_matrix(54) & cons_matrix(71) &
			cons_matrix(32) & cons_matrix(49) & cons_matrix(66) & cons_matrix(115) &
			cons_matrix(28) & cons_matrix(45) & cons_matrix(62) & cons_matrix(79) when sel_rd="001" else
			cons_matrix(40) & cons_matrix(57) & cons_matrix(74) & cons_matrix(123) &
			cons_matrix(36) & cons_matrix(53) & cons_matrix(70) & cons_matrix(87) &
			cons_matrix(48) & cons_matrix(65) & cons_matrix(82) & cons_matrix(3) &
			cons_matrix(44) & cons_matrix(61) & cons_matrix(78) & cons_matrix(95) &
			cons_matrix(56) & cons_matrix(73) & cons_matrix(90) & cons_matrix(11) &
			cons_matrix(52) & cons_matrix(69) & cons_matrix(86) & cons_matrix(103) &
			cons_matrix(64) & cons_matrix(81) & cons_matrix(98) & cons_matrix(19) &			
			cons_matrix(60) & cons_matrix(77) & cons_matrix(94) & cons_matrix(111) when sel_rd="010" else 
			cons_matrix(72) & cons_matrix(89) & cons_matrix(106) & cons_matrix(27) &
			cons_matrix(68) & cons_matrix(85) & cons_matrix(102) & cons_matrix(119) &
			cons_matrix(80) & cons_matrix(97) & cons_matrix(114) & cons_matrix(35) &
			cons_matrix(76) & cons_matrix(93) & cons_matrix(110) & cons_matrix(127) &
			cons_matrix(88) & cons_matrix(105) & cons_matrix(122) & cons_matrix(43) &
			cons_matrix(84) & cons_matrix(101) & cons_matrix(118) & cons_matrix(7) &
			cons_matrix(96) & cons_matrix(113) & cons_matrix(2) & cons_matrix(51) &
			cons_matrix(92) & cons_matrix(109) & cons_matrix(126) & cons_matrix(15) when sel_rd="100" else
			cons_matrix(104) & cons_matrix(121) & cons_matrix(10) & cons_matrix(59) &
			cons_matrix(100) & cons_matrix(117) & cons_matrix(6) & cons_matrix(23) &
			cons_matrix(112) & cons_matrix(1) & cons_matrix(18) & cons_matrix(67) &
			cons_matrix(108) & cons_matrix(125) & cons_matrix(14) & cons_matrix(31) &
			cons_matrix(120) & cons_matrix(9) & cons_matrix(26) & cons_matrix(75) &
			cons_matrix(116) & cons_matrix(5) & cons_matrix(22) & cons_matrix(39) &
			cons_matrix(0) & cons_matrix(17) & cons_matrix(34) & cons_matrix(83) &			
			cons_matrix(124) & cons_matrix(13) & cons_matrix(30) & cons_matrix(47);
	
end generate;
	
	
	sbox_gen: for i in 0 to n/(4*AES_SBOX_SIZE) - 1  generate
	sbox	: aes_sbox 	generic map (rom_style=>rom_style)
			port map (	 
				input=>to_sbox(AES_SBOX_SIZE*i + 7 downto AES_SBOX_SIZE*i), 
				output=>after_subbyte(AES_SBOX_SIZE*i+7 downto AES_SBOX_SIZE*i));	
	end generate;	

	after_shiftrow <= after_subbyte;

mc256: if n=GROESTL_DATA_SIZE_SMALL generate	
	mc			: entity work.groestl_mixbytes(groestl_mixbytes)	
				generic map (n=>n/4)
				port map (input=>after_shiftrow,  output=>after_mixcolumn);	
end generate;

mc512: if n=GROESTL_DATA_SIZE_BIG generate	
	mc_left			: entity work.groestl_mixbytes(groestl_mixbytes)	
					generic map (n=>n/4)
					port map (input=>after_shiftrow,  output=>after_mixcolumn);	

end generate;

	reg00		: regn generic map (n=>n/4, init=>zero) port map (clk=>clk, rst=>rst, en=>VCC, input=>after_mixcolumn, output=>after_rega);
	reg01		: regn generic map (n=>n/4, init=>zero) port map (clk=>clk, rst=>rst, en=>VCC, input=>after_rega, output=>after_regb);
	reg02		: regn generic map (n=>n/4, init=>zero) port map (clk=>clk, rst=>rst, en=>VCC, input=>after_regb, output=>after_regc);

	output <= after_regc & after_regb & after_rega & after_mixcolumn;
	
end q_folded_fx4;  



architecture p_folded_fx8 of groestl_pq_parallel_folded is

signal	after_subbyte		: std_logic_vector(n/8-1 downto 0);
signal	addcons			: std_logic_vector(n-1 downto 0);
signal	after_rega,after_regb,after_regc,
after_regd,after_rege,after_regf,after_regg: std_logic_vector(n/8-1 downto 0);
signal to_sbox			: std_logic_vector(n/8-1 downto 0);
signal	after_shiftrow	: std_logic_vector(n/8-1 downto 0);
constant zero			: std_logic_vector(n/8-1 downto 0):=(others=>'0');
signal after_mixcolumn	 : std_logic_vector(n/8-1 downto 0);
type matrix is array (0 to n/8) of std_logic_vector(7 downto 0);
signal cons_matrix	: matrix;

begin
	
	
gen256: if n=GROESTL_DATA_SIZE_SMALL generate

	addcons(511 downto 504) <= (input(511 downto 504) xor round xor x"00"); 
	addcons(503 downto 456) <= input(503 downto 456); 
	addcons(455 downto 448) <= input(455 downto 448); 
	
	addcons(447 downto 440) <= (input(447 downto 440) xor round xor x"10");
	addcons(439 downto 392) <= input(439 downto 392); 
	addcons(391 downto 384) <= input(391 downto 384); 

	addcons(383 downto 376) <= (input(383 downto 376) xor round xor x"20"); 
	addcons(375 downto 328) <= input(375 downto 328);  
	addcons(327 downto 320) <= input(327 downto 320); 

	addcons(319 downto 312) <= (input(319 downto 312) xor round xor x"30");
	addcons(311 downto 264) <= input(311 downto 264);  
	addcons(263 downto 256) <= input(263 downto 256);

	addcons(255 downto 248) <= (input(255 downto 248) xor round xor x"40");
	addcons(247 downto 200) <= input(247 downto 200);  
	addcons(199 downto 192) <= input(199 downto 192); 

	addcons(191 downto 184) <= (input(191 downto 184) xor round xor x"50"); 
	addcons(183 downto 136) <= input(183 downto 136);  
	addcons(135 downto 128) <= input(135 downto 128); 

	addcons(127 downto 120) <= (input(127 downto 120) xor round xor x"60"); 
	addcons(119 downto 72) <= input(119 downto 72); 
	addcons(71 downto 64) <= input(71 downto 64);

	addcons(63 downto 56) <= (input(63 downto 56) xor round xor x"70"); 
	addcons(55 downto 8) <= input(55 downto 8);  
	addcons(7 downto 0) <= input(7 downto 0); 
	
	cons256_gen: for i in 0 to GROESTL_DATA_SIZE_SMALL/8-1 generate
		cons_matrix(GROESTL_DATA_SIZE_SMALL/8-i-1) <= addcons(8*i+7 downto 8*i);
	end generate;
	
	to_sbox <= 	cons_matrix(0) & cons_matrix(9) & cons_matrix(18) & cons_matrix(27) &
				cons_matrix(36) & cons_matrix(45) & cons_matrix(54) & cons_matrix(63) when sel_rd="000" else
				cons_matrix(8) & cons_matrix(17) & cons_matrix(26) & cons_matrix(35) &
				cons_matrix(44) & cons_matrix(53) &	cons_matrix(62) & cons_matrix(7) when sel_rd="001" else
				cons_matrix(16) & cons_matrix(25) & cons_matrix(34) & cons_matrix(43) &
				cons_matrix(52) & cons_matrix(61) & cons_matrix(6) & cons_matrix(15) when sel_rd="010" else
				cons_matrix(24) & cons_matrix(33) & cons_matrix(42) & cons_matrix(51) &
				cons_matrix(60) & cons_matrix(5) & cons_matrix(14) & cons_matrix(23) when sel_rd="011" else 
				cons_matrix(32) & cons_matrix(41) & cons_matrix(50) & cons_matrix(59) &
				cons_matrix(4) & cons_matrix(13) & cons_matrix(22) & cons_matrix(31) when sel_rd="100" else
				cons_matrix(40) & cons_matrix(49) & cons_matrix(58) & cons_matrix(3) &
				cons_matrix(12) & cons_matrix(21) &	cons_matrix(30) & cons_matrix(39) when sel_rd="101" else
				cons_matrix(48) & cons_matrix(57) & cons_matrix(2) & cons_matrix(11) &
				cons_matrix(20) & cons_matrix(29) & cons_matrix(38) & cons_matrix(47) when sel_rd="110" else
				cons_matrix(56) & cons_matrix(1) & cons_matrix(10) & cons_matrix(19) &
				cons_matrix(28) & cons_matrix(37) & cons_matrix(46) & cons_matrix(55);				
				
				
end generate;			  		  

gen512: if n=GROESTL_DATA_SIZE_BIG generate	 

	addcons(1023 downto 1016) <= (input(1023 downto 1016) xor round xor x"00");
	addcons(1015 downto 968) <= input(1015 downto 968);
	addcons(967 downto 960) <= input(967 downto 960);

	addcons(959 downto 952) <= (input(959 downto 952) xor round xor x"10");
	addcons(951 downto 904) <= input(951 downto 904);
	addcons(903 downto 896) <= input(903 downto 896);

	addcons(895 downto 888) <= (input(895 downto 888) xor round xor x"20");
	addcons(887 downto 840) <= input(887 downto 840);
	addcons(839 downto 832) <= input(839 downto 832);

	addcons(831 downto 824) <= (input(831 downto 824) xor round xor x"30");
	addcons(823 downto 776) <= input(823 downto 776);
	addcons(775 downto 768) <= input(775 downto 768);

	addcons(767 downto 760) <= (input(767 downto 760) xor round xor x"40");
	addcons(759 downto 712) <= input(759 downto 712);
	addcons(711 downto 704) <= input(711 downto 704);

	addcons(703 downto 696) <= (input(703 downto 696) xor round xor x"50");
	addcons(695 downto 648) <= input(695 downto 648);
	addcons(647 downto 640) <= input(647 downto 640);

	addcons(639 downto 632) <= (input(639 downto 632) xor round xor x"60");
	addcons(631 downto 584) <= input(631 downto 584);
	addcons(583 downto 576) <= input(583 downto 576);

	addcons(575 downto 568) <= (input(575 downto 568) xor round xor x"70");
	addcons(567 downto 520) <= input(567 downto 520);
	addcons(519 downto 512) <= input(519 downto 512);

	addcons(511 downto 504) <= (input(511 downto 504) xor round xor x"80");
	addcons(503 downto 456) <= input(503 downto 456);
	addcons(455 downto 448) <= input(455 downto 448);
	
	addcons(447 downto 440) <= (input(447 downto 440) xor round xor x"90");
	addcons(439 downto 392) <= input(439 downto 392);
	addcons(391 downto 384) <= input(391 downto 384);

	addcons(383 downto 376) <= (input(383 downto 376) xor round xor x"a0");
	addcons(375 downto 328) <= input(375 downto 328);
	addcons(327 downto 320) <= input(327 downto 320);

	addcons(319 downto 312) <= (input(319 downto 312) xor round xor x"b0");
	addcons(311 downto 264) <= input(311 downto 264);
	addcons(263 downto 256) <= input(263 downto 256);

	addcons(255 downto 248) <= (input(255 downto 248) xor round xor x"c0");
	addcons(247 downto 200) <= input(247 downto 200);
	addcons(199 downto 192) <= input(199 downto 192);

	addcons(191 downto 184) <= (input(191 downto 184) xor round xor x"d0");
	addcons(183 downto 136) <= input(183 downto 136);
	addcons(135 downto 128) <= input(135 downto 128);

	addcons(127 downto 120) <= (input(127 downto 120) xor round xor x"e0");
	addcons(119 downto 72) <= input(119 downto 72);
	addcons(71 downto 64) <= input(71 downto 64);

	addcons(63 downto 56) <= (input(63 downto 56) xor round xor x"f0");
	addcons(55 downto 8) <= input(55 downto 8);
	addcons(7 downto 0) <= input(7 downto 0);
	
	cons512_gen: for i in 0 to GROESTL_DATA_SIZE_BIG/8-1 generate
		cons_matrix(GROESTL_DATA_SIZE_BIG/8-i-1) <= addcons(8*i+7 downto 8*i);
	end generate;
	
	to_sbox <= 	cons_matrix(0) & cons_matrix(9) & cons_matrix(18) & cons_matrix(27) &
			cons_matrix(36) & cons_matrix(45) & cons_matrix(54) & cons_matrix(95) &
			cons_matrix(8) & cons_matrix(17) & cons_matrix(26) & cons_matrix(35) &
			cons_matrix(44) & cons_matrix(53) &	cons_matrix(62) & cons_matrix(103) when sel_rd="000" else
			cons_matrix(16) & cons_matrix(25) & cons_matrix(34) & cons_matrix(43) &
			cons_matrix(52) & cons_matrix(61) & cons_matrix(70) & cons_matrix(111) &
			cons_matrix(24) & cons_matrix(33) & cons_matrix(42) & cons_matrix(51) &
			cons_matrix(60) & cons_matrix(69) & cons_matrix(78) & cons_matrix(119) when sel_rd="001" else
			cons_matrix(32) & cons_matrix(41) & cons_matrix(50) & cons_matrix(59) &
			cons_matrix(68) & cons_matrix(77) & cons_matrix(86) & cons_matrix(127) &
			cons_matrix(40) & cons_matrix(49) & cons_matrix(58) & cons_matrix(67) &
			cons_matrix(76) & cons_matrix(85) & cons_matrix(94) & cons_matrix(7) when sel_rd="010" else
			cons_matrix(48) & cons_matrix(57) & cons_matrix(66) & cons_matrix(75) &
			cons_matrix(84) & cons_matrix(93) & cons_matrix(102) & cons_matrix(15) &
			cons_matrix(56) & cons_matrix(65) & cons_matrix(74) & cons_matrix(83) &			
			cons_matrix(92) & cons_matrix(101) & cons_matrix(110) & cons_matrix(23) when sel_rd="011" else 
			cons_matrix(64) & cons_matrix(73) & cons_matrix(82) & cons_matrix(91) &
			cons_matrix(100) & cons_matrix(109) & cons_matrix(118) & cons_matrix(31) &
			cons_matrix(72) & cons_matrix(81) & cons_matrix(90) & cons_matrix(99) &
			cons_matrix(108) & cons_matrix(117) &	cons_matrix(126) & cons_matrix(39) when sel_rd="100" else
			cons_matrix(80) & cons_matrix(89) & cons_matrix(98) & cons_matrix(107) &
			cons_matrix(116) & cons_matrix(125) & cons_matrix(6) & cons_matrix(47) &
			cons_matrix(88) & cons_matrix(97) & cons_matrix(106) & cons_matrix(115) &
			cons_matrix(124) & cons_matrix(5) & cons_matrix(14) & cons_matrix(55) when sel_rd="101" else
			cons_matrix(96) & cons_matrix(105) & cons_matrix(114) & cons_matrix(123) &
			cons_matrix(4) & cons_matrix(13) & cons_matrix(22) & cons_matrix(63) &
			cons_matrix(104) & cons_matrix(113) & cons_matrix(122) & cons_matrix(3) &
			cons_matrix(12) & cons_matrix(21) & cons_matrix(30) & cons_matrix(71) when sel_rd="110" else
			cons_matrix(112) & cons_matrix(121) & cons_matrix(2) & cons_matrix(11) &
			cons_matrix(20) & cons_matrix(29) & cons_matrix(38) & cons_matrix(79) &
			cons_matrix(120) & cons_matrix(1) & cons_matrix(10) & cons_matrix(19) &			
			cons_matrix(28) & cons_matrix(37) & cons_matrix(46) & cons_matrix(87);	
			 
end generate;
		
	sbox_gen: for i in 0 to n/(8*AES_SBOX_SIZE) - 1  generate
	sbox	: aes_sbox 	generic map (rom_style=>rom_style)
			port map (	 
				input=>to_sbox(AES_SBOX_SIZE*i + 7 downto AES_SBOX_SIZE*i), 
				output=>after_subbyte(AES_SBOX_SIZE*i+7 downto AES_SBOX_SIZE*i));	
	end generate;	
	
	after_shiftrow <= after_subbyte;

mc256: if n=GROESTL_DATA_SIZE_SMALL generate	
	mc			: entity work.groestl_mixbytes(groestl_mixbytes)
				generic map (n=>n/8)
				port map (input=>after_shiftrow,  output=>after_mixcolumn);	
end generate;

mc512: if n=GROESTL_DATA_SIZE_BIG generate	
	mc_left			: entity work.groestl_mixbytes(groestl_mixbytes)
					generic map (n=>n/8)	
					port map (input=>after_shiftrow,  output=>after_mixcolumn);	
end generate;

	reg00		: regn generic map (n=>n/8, init=>zero) port map (clk=>clk, rst=>rst, en=>VCC, input=>after_mixcolumn, output=>after_rega);
	reg01		: regn generic map (n=>n/8, init=>zero) port map (clk=>clk, rst=>rst, en=>VCC, input=>after_rega, output=>after_regb);
	reg02		: regn generic map (n=>n/8, init=>zero) port map (clk=>clk, rst=>rst, en=>VCC, input=>after_regb, output=>after_regc);
	reg03		: regn generic map (n=>n/8, init=>zero) port map (clk=>clk, rst=>rst, en=>VCC, input=>after_regc, output=>after_regd);
	reg04		: regn generic map (n=>n/8, init=>zero) port map (clk=>clk, rst=>rst, en=>VCC, input=>after_regd, output=>after_rege);
	reg05		: regn generic map (n=>n/8, init=>zero) port map (clk=>clk, rst=>rst, en=>VCC, input=>after_rege, output=>after_regf);
	reg06		: regn generic map (n=>n/8, init=>zero) port map (clk=>clk, rst=>rst, en=>VCC, input=>after_regf, output=>after_regg);	
	
	output <= after_regg & after_regf & after_rege & after_regd & after_regc & after_regb & after_rega & after_mixcolumn;	
end p_folded_fx8; 


architecture q_folded_fx8 of groestl_pq_parallel_folded is

signal	after_subbyte		: std_logic_vector(n/8-1 downto 0);
signal	addcons			: std_logic_vector(n-1 downto 0);
signal	after_rega, after_regb, after_regc, 
after_regd, after_rege, after_regf, after_regg: std_logic_vector(n/8-1 downto 0);
signal	after_shiftrow	: std_logic_vector(n/8-1 downto 0);
constant zero			: std_logic_vector(n/8-1 downto 0):=(others=>'0'); 
signal after_mixcolumn, to_sbox	 : std_logic_vector(n/8-1 downto 0);
type matrix is array (0 to n/8-1) of std_logic_vector(7 downto 0);
signal cons_matrix	: matrix;
begin
	
	
gen256: if n=GROESTL_DATA_SIZE_SMALL generate

	addcons(511 downto 504) <=  (not input(511 downto 504));
	addcons(503 downto 456) <=  (not input(503 downto 456)); 
	addcons(455 downto 448) <=  (x"ff" xor input(455 downto 448) xor round);
	
	addcons(447 downto 440) <=  (not input(447 downto 440));
	addcons(439 downto 392) <=  (not input(439 downto 392)); 
	addcons(391 downto 384) <=  (x"ef" xor input(391 downto 384) xor round);

	addcons(383 downto 376) <=  (not input(383 downto 376));
	addcons(375 downto 328) <= (not input(375 downto 328)); 
	addcons(327 downto 320) <= (x"df" xor input(327 downto 320) xor round);

	addcons(319 downto 312) <=  (not input(319 downto 312));
	addcons(311 downto 264) <= (not input(311 downto 264)); 
	addcons(263 downto 256) <= (x"cf" xor input(263 downto 256) xor round);

	addcons(255 downto 248) <=  (not input(255 downto 248));
	addcons(247 downto 200) <=  (not input(247 downto 200)); 
	addcons(199 downto 192) <=  (x"bf" xor input(199 downto 192) xor round);

	addcons(191 downto 184) <=  (not input(191 downto 184));
	addcons(183 downto 136) <=  (not input(183 downto 136)); 
	addcons(135 downto 128) <=  (x"af" xor input(135 downto 128) xor round);

	addcons(127 downto 120) <=  (not input(127 downto 120));
	addcons(119 downto 72) <=  (not input(119 downto 72)); 
	addcons(71 downto 64) <=  (x"9f" xor input(71 downto 64) xor round);

	addcons(63 downto 56) <=  (not input(63 downto 56));
	addcons(55 downto 8) <=  (not input(55 downto 8)); 
	addcons(7 downto 0) <=  (x"8f" xor input(7 downto 0) xor round);
		
	cons256_gen: for i in 0 to GROESTL_DATA_SIZE_SMALL/8-1 generate
		cons_matrix(GROESTL_DATA_SIZE_SMALL/8-i-1) <= addcons(8*i+7 downto 8*i);
	end generate;
	
		to_sbox <= 	cons_matrix(8) & cons_matrix(25) & cons_matrix(42) & cons_matrix(59) &
				cons_matrix(4) & cons_matrix(21) & cons_matrix(38) & cons_matrix(55) when sel_rd="000" else
				cons_matrix(16) & cons_matrix(33) & cons_matrix(50) & cons_matrix(3) &
				cons_matrix(12) & cons_matrix(29) &	cons_matrix(46) & cons_matrix(63) when sel_rd="001" else
				cons_matrix(24) & cons_matrix(41) & cons_matrix(58) & cons_matrix(11) &
				cons_matrix(20) & cons_matrix(37) & cons_matrix(54) & cons_matrix(7) when sel_rd="010" else
				cons_matrix(32) & cons_matrix(49) & cons_matrix(2) & cons_matrix(19) &
				cons_matrix(28) & cons_matrix(45) & cons_matrix(62) & cons_matrix(15) when sel_rd="011" else 
				cons_matrix(40) & cons_matrix(57) & cons_matrix(10) & cons_matrix(27) &
				cons_matrix(36) & cons_matrix(53) & cons_matrix(6) & cons_matrix(23) when sel_rd="100" else
				cons_matrix(48) & cons_matrix(1) & cons_matrix(18) & cons_matrix(35) &
				cons_matrix(44) & cons_matrix(61) &	cons_matrix(14) & cons_matrix(31) when sel_rd="101" else
				cons_matrix(56) & cons_matrix(9) & cons_matrix(26) & cons_matrix(43) &
				cons_matrix(52) & cons_matrix(5) & cons_matrix(22) & cons_matrix(39) when sel_rd="110" else
				cons_matrix(0) & cons_matrix(17) & cons_matrix(34) & cons_matrix(51) &
				cons_matrix(60) & cons_matrix(13) & cons_matrix(30) & cons_matrix(47);	
	
end generate;			  		  

gen512: if n=GROESTL_DATA_SIZE_BIG generate	 

	addcons(1023 downto 1016) <=  (not input(1023 downto 1016));
	addcons(1015 downto 968) <=  (not input(1015 downto 968)); 
	addcons(967 downto 960) <=  (x"ff" xor input(967 downto 960) xor round); 

	addcons(959 downto 952) <=  (not input(959 downto 952));
	addcons(951 downto 904) <=  (not input(951 downto 904)); 
	addcons(903 downto 896) <=  (x"ef" xor input(903 downto 896) xor round); 

	addcons(895 downto 888) <=  (not input(895 downto 888));
	addcons(887 downto 840) <=  (not input(887 downto 840)); 
	addcons(839 downto 832) <=  (x"df" xor input(839 downto 832) xor round);  

	addcons(831 downto 824) <=  (not input(831 downto 824));
	addcons(823 downto 776) <=  (not input(823 downto 776)); 
	addcons(775 downto 768) <=  (x"cf" xor input(775 downto 768) xor round);

	addcons(767 downto 760) <=  (not input(767 downto 760));
	addcons(759 downto 712) <=  (not input(759 downto 712)); 
	addcons(711 downto 704) <=  (x"bf" xor input(711 downto 704) xor round);

	addcons(703 downto 696) <=  (not input(703 downto 696));
	addcons(695 downto 648) <=  (not input(695 downto 648)); 
	addcons(647 downto 640) <=  (x"af" xor input(647 downto 640) xor round);

	addcons(639 downto 632) <=  (not input(639 downto 632));
	addcons(631 downto 584) <=  (not input(631 downto 584)); 
	addcons(583 downto 576) <=  (x"9f" xor input(583 downto 576) xor round);

	addcons(575 downto 568) <=  (not input(575 downto 568));
	addcons(567 downto 520) <= (not input(567 downto 520)); 
	addcons(519 downto 512) <=  (x"8f" xor input(519 downto 512) xor round);

	addcons(511 downto 504) <=  (not input(511 downto 504));
	addcons(503 downto 456) <=  (not input(503 downto 456)); 
	addcons(455 downto 448) <=  (x"7f" xor input(455 downto 448) xor round);
	
	addcons(447 downto 440) <=  (not input(447 downto 440));
	addcons(439 downto 392) <=  (not input(439 downto 392)); 
	addcons(391 downto 384) <=  (x"6f" xor input(391 downto 384) xor round);

	addcons(383 downto 376) <=  (not input(383 downto 376));
	addcons(375 downto 328) <=  (not input(375 downto 328)); 
	addcons(327 downto 320) <=  (x"5f" xor input(327 downto 320) xor round);

	addcons(319 downto 312) <=  (not input(319 downto 312));
	addcons(311 downto 264) <=  (not input(311 downto 264)); 
	addcons(263 downto 256) <=  (x"4f" xor input(263 downto 256) xor round);

	addcons(255 downto 248) <=  (not input(255 downto 248));
	addcons(247 downto 200) <=  (not input(247 downto 200)); 
	addcons(199 downto 192) <=  (x"3f" xor input(199 downto 192) xor round);

	addcons(191 downto 184) <=  (not input(191 downto 184));
	addcons(183 downto 136) <=  (not input(183 downto 136)); 
	addcons(135 downto 128) <=  (x"2f" xor input(135 downto 128) xor round);

	addcons(127 downto 120) <=  (not input(127 downto 120));
	addcons(119 downto 72) <=  (not input(119 downto 72)); 
	addcons(71 downto 64) <=  (x"1f" xor input(71 downto 64) xor round);

	addcons(63 downto 56) <=  (not input(63 downto 56));
	addcons(55 downto 8) <=  (not input(55 downto 8)); 
	addcons(7 downto 0) <=  (x"0f" xor input(7 downto 0) xor round);

	cons512_gen: for i in 0 to GROESTL_DATA_SIZE_BIG/8-1 generate
		cons_matrix(GROESTL_DATA_SIZE_BIG/8-i-1) <= addcons(8*i+7 downto 8*i);
	end generate;
	
	to_sbox <= 	cons_matrix(8) & cons_matrix(25) & cons_matrix(42) & cons_matrix(91) &
			cons_matrix(4) & cons_matrix(21) & cons_matrix(38) & cons_matrix(55) &
			cons_matrix(16) & cons_matrix(33) & cons_matrix(50) & cons_matrix(99) &
			cons_matrix(12) & cons_matrix(29) &	cons_matrix(46) & cons_matrix(63) when sel_rd="000" else
			cons_matrix(24) & cons_matrix(41) & cons_matrix(58) & cons_matrix(107) &
			cons_matrix(20) & cons_matrix(37) & cons_matrix(54) & cons_matrix(71) &
			cons_matrix(32) & cons_matrix(49) & cons_matrix(66) & cons_matrix(115) &
			cons_matrix(28) & cons_matrix(45) & cons_matrix(62) & cons_matrix(79) when sel_rd="001" else
			cons_matrix(40) & cons_matrix(57) & cons_matrix(74) & cons_matrix(123) &
			cons_matrix(36) & cons_matrix(53) & cons_matrix(70) & cons_matrix(87) &
			cons_matrix(48) & cons_matrix(65) & cons_matrix(82) & cons_matrix(3) &
			cons_matrix(44) & cons_matrix(61) & cons_matrix(78) & cons_matrix(95) when sel_rd="010" else
			cons_matrix(56) & cons_matrix(73) & cons_matrix(90) & cons_matrix(11) &
			cons_matrix(52) & cons_matrix(69) & cons_matrix(86) & cons_matrix(103) &
			cons_matrix(64) & cons_matrix(81) & cons_matrix(98) & cons_matrix(19) &			
			cons_matrix(60) & cons_matrix(77) & cons_matrix(94) & cons_matrix(111) when sel_rd="011" else 
			cons_matrix(72) & cons_matrix(89) & cons_matrix(106) & cons_matrix(27) &
			cons_matrix(68) & cons_matrix(85) & cons_matrix(102) & cons_matrix(119) &
			cons_matrix(80) & cons_matrix(97) & cons_matrix(114) & cons_matrix(35) &
			cons_matrix(76) & cons_matrix(93) & cons_matrix(110) & cons_matrix(127) when sel_rd="100" else
			cons_matrix(88) & cons_matrix(105) & cons_matrix(122) & cons_matrix(43) &
			cons_matrix(84) & cons_matrix(101) & cons_matrix(118) & cons_matrix(7) &
			cons_matrix(96) & cons_matrix(113) & cons_matrix(2) & cons_matrix(51) &
			cons_matrix(92) & cons_matrix(109) & cons_matrix(126) & cons_matrix(15) when sel_rd="101" else
			cons_matrix(104) & cons_matrix(121) & cons_matrix(10) & cons_matrix(59) &
			cons_matrix(100) & cons_matrix(117) & cons_matrix(6) & cons_matrix(23) &
			cons_matrix(112) & cons_matrix(1) & cons_matrix(18) & cons_matrix(67) &
			cons_matrix(108) & cons_matrix(125) & cons_matrix(14) & cons_matrix(31) when sel_rd="110" else
			cons_matrix(120) & cons_matrix(9) & cons_matrix(26) & cons_matrix(75) &
			cons_matrix(116) & cons_matrix(5) & cons_matrix(22) & cons_matrix(39) &
			cons_matrix(0) & cons_matrix(17) & cons_matrix(34) & cons_matrix(83) &			
			cons_matrix(124) & cons_matrix(13) & cons_matrix(30) & cons_matrix(47);
	
end generate;
	
	
	sbox_gen: for i in 0 to n/(8*AES_SBOX_SIZE) - 1  generate
	sbox	: aes_sbox 	generic map (rom_style=>rom_style)
			port map (	 
				input=>to_sbox(AES_SBOX_SIZE*i + 7 downto AES_SBOX_SIZE*i), 
				output=>after_subbyte(AES_SBOX_SIZE*i+7 downto AES_SBOX_SIZE*i));	
	end generate;	

	after_shiftrow <= after_subbyte;

mc256: if n=GROESTL_DATA_SIZE_SMALL generate	
	mc			: entity work.groestl_mixbytes(groestl_mixbytes)	
				generic map (n=>n/8)
				port map (input=>after_shiftrow,  output=>after_mixcolumn);	
end generate;

mc512: if n=GROESTL_DATA_SIZE_BIG generate	
	mc_left			: entity work.groestl_mixbytes(groestl_mixbytes)	
					generic map (n=>n/8)
					port map (input=>after_shiftrow,  output=>after_mixcolumn);	

end generate;

	reg00		: regn generic map (n=>n/8, init=>zero) port map (clk=>clk, rst=>rst, en=>VCC, input=>after_mixcolumn, output=>after_rega);
	reg01		: regn generic map (n=>n/8, init=>zero) port map (clk=>clk, rst=>rst, en=>VCC, input=>after_rega, output=>after_regb);
	reg02		: regn generic map (n=>n/8, init=>zero) port map (clk=>clk, rst=>rst, en=>VCC, input=>after_regb, output=>after_regc);
	reg03		: regn generic map (n=>n/8, init=>zero) port map (clk=>clk, rst=>rst, en=>VCC, input=>after_regc, output=>after_regd);
	reg04		: regn generic map (n=>n/8, init=>zero) port map (clk=>clk, rst=>rst, en=>VCC, input=>after_regd, output=>after_rege);
	reg05		: regn generic map (n=>n/8, init=>zero) port map (clk=>clk, rst=>rst, en=>VCC, input=>after_rege, output=>after_regf);
	reg06		: regn generic map (n=>n/8, init=>zero) port map (clk=>clk, rst=>rst, en=>VCC, input=>after_regf, output=>after_regg);	
	
	output <= after_regg & after_regf & after_rege & after_regd & after_regc & after_regb & after_rega & after_mixcolumn;	
	
end q_folded_fx8; 